/*
  Desktop Station Empty Box
  Copyright(C) '2016- Desktop Station / yaasan

  The circuit:
 * LCD RS pin to digital pin 9
 * LCD Enable pin to digital pin 8
 * LCD D4 pin to digital pin 13
 * LCD D5 pin to digital pin 12
 * LCD D6 pin to digital pin 11
 * LCD D7 pin to digital pin 10
 * LCD R/W pin to ground
 * LCD VSS pin to ground
 * LCD VCC pin to 5V
 * 10K resistor:
 * ends to +5V and ground
 * wiper to LCD VO pin (pin 3)

 This example code is in the public domain.

 http://www.arduino.cc/en/Tutorial/LiquidCrystal
 */

// include the library code:
#include <LiquidCrystal.h>
#include <EEPROM.h>
#include <avr/wdt.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "DSSequence.h"
#include "ManufactureList.h"

const char rMainMenu0[] PROGMEM    = " CV Read";
const char rMainMenu1[] PROGMEM    = " CV Write";
const char rMainMenu2[] PROGMEM    = " CV ReadWrite";
const char rMainMenu3[] PROGMEM    = " Nucky Signal";
const char rMainMenu4[] PROGMEM    = " Check LocAddr";
const char rMainMenu5[] PROGMEM    = " Write LocAddr";
const char rMainMenu6[] PROGMEM    = " Loc Control";
const char rMainMenu7[] PROGMEM    = " Acc Control";
const char rMainMenu8[] PROGMEM    = " Config";
const char rManCheck[] PROGMEM     = " Manufacturer";
const char rCVWriting[] PROGMEM    = " CV Writing...";
const char rCVReading[] PROGMEM    = " CV Reading...";
const char rCVEnterStart[] PROGMEM   = " Enter to start";
const char rCVReadErr[] PROGMEM      = " Read error!";
const char rLocAdrWrite[] PROGMEM    = " LocAddr Write";
const char rLocAdrCheck[] PROGMEM    = " LocAddr Check";
const char rWriteOk[] PROGMEM     = " Write Ok!";
const char rWriteEnd[] PROGMEM    = " Write End!";
const char rWriteErr[] PROGMEM    = " Write error!";
const char rFinished[] PROGMEM    = " Finished!";
const char rFailed[] PROGMEM      = " Failed!";
const char rReturnTop[] PROGMEM      = " Return to Top";


// initialize the library with the numbers of the interface pins
LiquidCrystal lcd(9, 8, 13, 12, 11, 10);

DSSequence Sequence;
DSCoreLib DSCore;

//Task Schedule
unsigned long gPreviousL5 = 0; //   500ms interval
unsigned long gPreviousL1 = 0; //   100ms interval

//LCD text
char gTextBuf[16 + 1];
byte gDisplayUpdate = 0;

void displayConfig(byte inNo, byte inArrow);
void displayCVRead(byte inNo);
void displayCVReadWrite(byte inNo);
void displayAddrCheck(byte inNo);
void displayAddrWrite(byte inNo);
void displayNuckySignal(byte inNo);
void ExecuteCommand(unsigned char *outpReturn, unsigned char inDrawFlag);

void changedPowerStatus(const byte inPower);
void changedTargetSpeed(const word inAddr, const word inSpeed);
void changedTargetFunc(const word inAddr, const byte inFuncNo, const byte inFuncPower);
void changedTargetDir(const word inAddr, const byte inDir);
void changedTurnout(const word inAddr, const byte inDir);
void clearTextBuf(void);

void setup() 
{
	
	DSCore.Init();
	
	// set up the LCD's number of columns and rows:
	lcd.begin(16, 2);
	// Print a message to the LCD.
	lcd.print("Desktop Station");
	lcd.setCursor(0, 1);
	lcd.print("CVRead&Writer 4e");
	
	
	Serial.begin(115200);
	while (!Serial);

	Serial.println("Desktop Station Empty Box 4e");
	Serial.println("2017 Apr 21");
	
	delay(1000);
	
	//DCCs֐
	Sequence.CVExeFunc = ExecuteCommand;
	Sequence.ChangedTargetSpeed = changedTargetSpeed;
	Sequence.ChangedTargetFunc = changedTargetFunc;
	Sequence.ChangedTargetDir = changedTargetDir;
	Sequence.ChangedTurnout = changedTurnout;
	Sequence.ChangedPowerStatus = changedPowerStatus;
	/* EHb`hbOL(8sec) */
	wdt_enable(WDTO_8S); // WDT reset
	
	
	//lcd update
	gDisplayUpdate = 0;
	Draw();
	
	
	gPreviousL5 = millis();
	gPreviousL1 = millis();
	sei();    //ENABLE INTERRUPTION
}

void loop() 
{
	
	
	/* Reset watchdog timer. */
	wdt_reset(); 
	
	/* DSCore Scan */
	if( (Sequence.gMode == MODE_LOCCTRL) || (Sequence.gMode == MODE_ACCCTRL))
	{
		DSCore.Scan();
	}
	
	//Button monitoring interval
	if( (millis() - gPreviousL1) >= 100)
	{
		
		gDisplayUpdate = Sequence.Interval();
		
		//Reset task
		gPreviousL1 = millis();
	}
	
	// LCD update interval
//	if( ( (millis() - gPreviousL5) >= 500) || ( gDisplayUpdate > 0))
	if( gDisplayUpdate > 0 )
	{
		gDisplayUpdate = 0;
		Draw();
		
		
		//Reset task
		//gPreviousL5 = millis();
	}
	
}



void Draw()
{
	lcd.clear();
	
	switch(Sequence.gMode)
	{
	case MODE_MAINMENU:
		
		for( int i = 0; i <= 1; i++)
		{
			lcd.setCursor(0, i);
			displayMainMenu(i + Sequence.gNo_MainMenu, i);
			
		}
		
		break;
		
	case MODE_CVREAD:
		
		displayCVRead(Sequence.gNo_CVReadMenu);
		
		
		break;
		
	case MODE_CVWRITE:
		
		displayCVWrite(Sequence.gNo_CVWriteMenu);
		
		
		break;
		
	case MODE_CVREADWRITE:
		
		displayCVReadWrite(Sequence.gNo_CVReadWriteMenu);
		
		
		break;
	
	case MODE_CHECKADDRESS:
		displayAddrCheck(Sequence.gNo_AddrChkMenu);
		
		break;
		
	case MODE_WRITEADDRESS:
		displayAddrWrite(Sequence.gNo_AddrWriteMenu);
		
		break;
		
	case MODE_NUCKYSIGNAL:
		displayNuckySignal(Sequence.gNo_NSignalMenu);
		
		break;
		
	case MODE_LOCCTRL:
		displayLocCtrl(Sequence.gNo_LocCtrlMenu);
		
		break;
	case MODE_ACCCTRL:
		displayAccCtrl(Sequence.gNo_AccCtrlMenu);
		break;
		
	case MODE_MANUFACTURE:
		displayManufacture(Sequence.gNo_ManCtrlMenu);
		break;
		
	case MODE_CONFIG:
		
		for( int i = 0; i <= 1; i++)
		{
			lcd.setCursor(0, i);
			displayConfig(i + Sequence.gNo_ConfigMenu, i);
			
		}
		break;
		
	}
	
	
	
}


void displayConfig(byte inNo, byte inArrow)
{
	char aTextBuf[16 + 1] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	switch(inNo)
	{
	case 0:
		sprintf(aTextBuf, " Read mode: %s",  (Sequence.gCfgReadSpeed == 0) ? "Fast" : "Slow");
		break;
	case 1:
		StrCpyFromROM(aTextBuf, rReturnTop, 14);
		//sprintf(aTextBuf, " Return to Top");
		break;
	}
	
	if( inArrow == 0)
	{
		aTextBuf[0] = 0x3E;
	}
	
	lcd.print(aTextBuf);
	
}


void displayCVRead(byte inNo)
{
	char aTextBuf[16 + 1] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	//1s
	sprintf(aTextBuf, " CV No.  %04d", Sequence.CVNo);
	
	if( inNo == 0)
	{
		aTextBuf[0] = 0x3E;
	}
	lcd.setCursor(0, 0);
	lcd.print(aTextBuf);
	
	//Next row
	lcd.setCursor(0, 1);
	
	if( inNo == 1)
	{
	
		//2s
		StrCpyFromROM(aTextBuf, rCVReading, 14);
		//sprintf(aTextBuf, " CV Reading...");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 2)
	{
	
		//2s
		sprintf(aTextBuf, " CV Value %03d", Sequence.CVValue);
		
		aTextBuf[0] = 0x3E;
		
		lcd.print(aTextBuf);
	
	}
	else if (inNo == 3)
	{
		
		StrCpyFromROM(aTextBuf, rCVReadErr, 12);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
	
}

void displayCVWrite(byte inNo)
{
	char aTextBuf[16 + 1] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	//1s
	sprintf(aTextBuf, " CV No.  %04d", Sequence.CVNo);
	
	if( inNo == 0)
	{
		aTextBuf[0] = 0x3E;
	}
	lcd.setCursor(0, 0);
	lcd.print(aTextBuf);
	
	//Next row
	lcd.setCursor(0, 1);
	
	if( inNo <= 1)
	{
	
		//2s
		sprintf(aTextBuf, " CV Value %03d", Sequence.CVValue);
		
		if( inNo == 1)
		{
			aTextBuf[0] = 0x3E;
		}
		
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 2)
	{
		
		StrCpyFromROM(aTextBuf, rCVWriting, 14);
		//sprintf(aTextBuf, " CV Writing...");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	}
	else if (inNo == 3)
	{
		
		StrCpyFromROM(aTextBuf, rWriteOk, 10);
		//sprintf(aTextBuf, " Write Ok!");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
	else if (inNo == 4)
	{
		
		StrCpyFromROM(aTextBuf, rWriteErr, 13);
		//sprintf(aTextBuf, " Write error!");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
	
}


void displayCVReadWrite(byte inNo)
{
	char aTextBuf[16 + 1] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	//1s
	sprintf(aTextBuf, " CV No.  %04d", Sequence.CVNo);
	
	if( inNo == 0)
	{
		aTextBuf[0] = 0x3E;
	}
	lcd.setCursor(0, 0);
	lcd.print(aTextBuf);
	
	//Next row
	lcd.setCursor(0, 1);
	
	if( inNo == 1)
	{
	
		//2s
		StrCpyFromROM(aTextBuf, rCVReading, 14);
		//sprintf(aTextBuf, " CV Reading...");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 2)
	{
	
		//2s
		sprintf(aTextBuf, " CV Value %03d", Sequence.CVValue);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 3)
	{
		
		StrCpyFromROM(aTextBuf, rCVWriting, 14);
		//sprintf(aTextBuf, " CV Writing...");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	}
	else if (inNo == 4)
	{
		
		StrCpyFromROM(aTextBuf, rWriteOk, 10);
		//sprintf(aTextBuf, " Write OK!");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
	else if (inNo == 5)
	{
		
		StrCpyFromROM(aTextBuf, rWriteErr, 13);
		//sprintf(aTextBuf, " Write error!");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
	else if (inNo == 6)
	{
		
		StrCpyFromROM(aTextBuf, rCVReadErr, 12);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
}


void displayAddrCheck(byte inNo)
{
	char aTextBuf[16 + 1] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	//1s
	StrCpyFromROM(aTextBuf, rLocAdrCheck, 14);
	lcd.setCursor(0, 0);
	lcd.print(aTextBuf);
	
	//Next row
	lcd.setCursor(0, 1);
	
	if( inNo == 0)
	{
	
		//2s
		StrCpyFromROM(aTextBuf, rCVEnterStart, 15);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 1)
	{
	
		//2s
		StrCpyFromROM(aTextBuf, rCVReading, 14);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 2)
	{
	
		//2s
		sprintf(aTextBuf, " %04d(%s)", Sequence.Address, (Sequence.Address > 100) ? "LONG" : "SHORT");
		
		aTextBuf[0] = 0x3E;
		
		lcd.print(aTextBuf);
	
	}
	else if (inNo == 99)
	{
		
		StrCpyFromROM(aTextBuf, rCVReadErr, 12);
		//sprintf(aTextBuf, " Read error!");
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
	
}


void displayAddrWrite(byte inNo)
{
	char aTextBuf[16 + 1] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	//1s
	StrCpyFromROM(aTextBuf, rLocAdrWrite, 14);
	lcd.setCursor(0, 0);
	lcd.print(aTextBuf);
	
	//Next row
	lcd.setCursor(0, 1);
	
	if( inNo == 0)
	{
	
		//2s
		sprintf(aTextBuf, " %04d", Sequence.Address);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 1)
	{
	
		//2s
		StrCpyFromROM(aTextBuf, rCVReading, 14);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 2)
	{
	
		//2s
		StrCpyFromROM(aTextBuf, rCVWriting, 14);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 3)
	{
	
		//2s
		StrCpyFromROM(aTextBuf, rFinished, 10);
		
		aTextBuf[0] = 0x3E;
		
		lcd.print(aTextBuf);
	
	}
	else if( inNo == 4)
	{
	
		//2s
		StrCpyFromROM(aTextBuf, rFailed, 8);
		
		aTextBuf[0] = 0x3E;
		
		lcd.print(aTextBuf);
	
	}
	else if (inNo == 99)
	{
		
		StrCpyFromROM(aTextBuf, rCVReadErr, 12);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
	
}

void displayNuckySignal(byte inNo)
{
	char aTextBuf[16 + 1] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	if( inNo <= 5)
	{
	
		//1s
		sprintf(aTextBuf, " T%03d A%03d B%03d", Sequence.SignalAddr[0], Sequence.SignalAddr[1], Sequence.SignalAddr[2]);
		
		switch(inNo)
		{
			case 0: aTextBuf[0] = 0x3E; break;
			case 1: aTextBuf[5] = 0x3E; break;
			case 2: aTextBuf[10] = 0x3E; break;
			case 3: break;
			case 4: break;
			case 5: break;
		}
		
		lcd.setCursor(0, 0);
		lcd.print(aTextBuf);
		
		//Clear
		for( int i = 0; i < 16; i++)
		{
			aTextBuf[i] = 0;
		}
		
		
		//2s
		sprintf(aTextBuf, " C%03d D%03d E%03d", Sequence.SignalAddr[3], Sequence.SignalAddr[4], Sequence.SignalAddr[5]);
		
		switch(inNo)
		{
			case 0: break;
			case 1: break;
			case 2: break;
			case 3: aTextBuf[0] = 0x3E;break;
			case 4: aTextBuf[5] = 0x3E;break;
			case 5: aTextBuf[10] = 0x3E;break;
		}
		
		//Next row
		lcd.setCursor(0, 1);
		lcd.print(aTextBuf);
	
	}
	else
	{
		StrCpyFromROM(aTextBuf, rMainMenu3, 13);
		lcd.setCursor(0, 0);
		lcd.print(aTextBuf);
		
		if( inNo == 6)
		{
			StrCpyFromROM(aTextBuf, rCVWriting, 14);
			
		}
		else if (inNo == 7)
		{
			StrCpyFromROM(aTextBuf, rWriteOk, 10);
			
		}
		else if (inNo == 99)
		{
			StrCpyFromROM(aTextBuf, rWriteEnd, 11);
			
		}

		//Next row
		lcd.setCursor(0, 1);
		aTextBuf[0] = 0x3E;
		lcd.print(aTextBuf);
		
	}
}


void displayMainMenu(byte inNo, byte inArrow)
{
	char aTextBuf[16 + 1] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	
	switch(inNo)
	{
	case 0:
		StrCpyFromROM(aTextBuf, rMainMenu0, 8);
		break;
	case 1:
		StrCpyFromROM(aTextBuf, rMainMenu1, 9);
		break;
	case 2:
		StrCpyFromROM(aTextBuf, rMainMenu2, 13);
		break;
	case 3:
		StrCpyFromROM(aTextBuf, rMainMenu3, 13);
		break;
	case 4:
		StrCpyFromROM(aTextBuf, rMainMenu4, 14);
		break;
	case 5:
		StrCpyFromROM(aTextBuf, rMainMenu5, 14);
		break;
	case 6:
		StrCpyFromROM(aTextBuf, rMainMenu6, 12);
		break;
	case 7:
		StrCpyFromROM(aTextBuf, rMainMenu7, 12);
		break;
	case 8:
		StrCpyFromROM(aTextBuf, rManCheck, 13);
		break;
	case 9:
		StrCpyFromROM(aTextBuf, rMainMenu8, 7);
		break;
	}
	
	if( inArrow == 0)
	{
		aTextBuf[0] = 0x3E;
	}
	
	lcd.print(aTextBuf);
	
	
}

char getCh(uint32_t inFuncNo)
{
	char aChr = '_';
	
	if( ((Sequence.gLoc_Func >> inFuncNo) & 1) > 0)
	{
		
		aChr = '*';
	}
	
	return aChr;
}

void displayLocCtrl(byte inNo)
{
	clearTextBuf();
	
	sprintf(gTextBuf, " Loc%04d Spd%01d%s", Sequence.gLoc_Addr, Sequence.gLoc_Spd, Sequence.gLoc_Dir == 0 ? " F>" : " R<");
	
	//\ʒu
	switch(inNo)
	{
	case 0:
		gTextBuf[0] = 0x3E;
		break;
	case 1:
		gTextBuf[8] = 0x3E;
		break;
	}
	
	
	//Next row
	lcd.setCursor(0, 0);
	lcd.print(gTextBuf);
	
	//NA
	clearTextBuf();
	
	switch(inNo)
	{
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
		sprintf(gTextBuf, " F00%c F01%c F02%c ", getCh(0), getCh(1), getCh(2));
		break;
	case 5:
	case 6:
	case 7:
		sprintf(gTextBuf, " F03%c F04%c F05%c ", getCh(3), getCh(4), getCh(5));
		break;
	case 8:
	case 9:
	case 10:
		sprintf(gTextBuf, " F06%c F07%c F08%c ", getCh(6), getCh(7), getCh(8));
		break;
	case 11:
	case 12:
	case 13:
		sprintf(gTextBuf, " F09%c F10%c F11%c", getCh(9), getCh(10), getCh(11));
		break;
	case 14:
	case 15:
	case 16:
		sprintf(gTextBuf, " F12%c F13%c F14%c", getCh(12), getCh(13), getCh(14));
		break;
	case 17:
	case 18:
	case 19:
		sprintf(gTextBuf, " F15%c F16%c F17%c", getCh(15), getCh(16), getCh(17));
		break;
	case 20:
	case 21:
	case 22:
		sprintf(gTextBuf, " F18%c F19%c F20%c", getCh(18), getCh(19), getCh(20));
		break;
	case 23:
	case 24:
	case 25:
		sprintf(gTextBuf, " F21%c F22%c F23%c", getCh(21), getCh(22), getCh(23));
		break;
	case 26:
	case 27:
	case 28:
		sprintf(gTextBuf, " F24%c F25%c F26%c ", getCh(24), getCh(25), getCh(26));
		break;
	case 29:
	case 30:
		sprintf(gTextBuf, " F27%c F28%c       ", getCh(27), getCh(28));
		break;
	}
	
	//̃Zbg
	switch(inNo)
	{
	case 2:
	case 5:
	case 8:
	case 11:
	case 14:
	case 17:
	case 20:
	case 23:
	case 26:
	case 29:
		gTextBuf[0] = 0x3E;
		break;
		
	case 3:
	case 6:
	case 9:
	case 12:
	case 15:
	case 18:
	case 21:
	case 24:
	case 27:
	case 30:
		gTextBuf[5] = 0x3E;
		break;
		
	case 4:
	case 7:
	case 10:
	case 13:
	case 16:
	case 19:
	case 22:
	case 25:
	case 28:
		gTextBuf[10] = 0x3E;
		break;

	default:
		break;
	}
	
	//Next row
	lcd.setCursor(0, 1);
	lcd.print(gTextBuf);
	
	
}

void displayAccCtrl(byte inNo)
{
	clearTextBuf();
	
	if( Sequence.gAcc_Addr == 0)
	{
		sprintf(gTextBuf, " Acc0000 ----");
	}
	else
	{
		sprintf(gTextBuf, " Acc%04d %s", Sequence.gAcc_Addr,  Sequence.gAcc_Dir == 1 ? "STR|" : "DIV/");
	}
	
	//Next row
	lcd.setCursor(0, 0);
	
	lcd.print(gTextBuf);
	
	//Next row
	lcd.setCursor(0, 1);
	clearTextBuf();
	
	sprintf(gTextBuf, " %04d ", (Sequence.gAcc_Addr & 0xFFF8));

	for( int i = 0; i < 8; i++)
	{
		if( ((Sequence.gAcc_FuncBuf[Sequence.gAcc_Addr >> 3] >> i) & 1) == 0)
		{
			gTextBuf[i + 6] = 0x5f;
		}
		else
		{
			gTextBuf[i + 6] = 0x2a;
		}
	}
	
	lcd.print(gTextBuf);
	

}

void clearTextBuf(void)
{
	for( int i = 0; i < 16; i++)
	{
		gTextBuf[i] = 0x00;
	}
	
}

void displayManufacture(uint8_t inNo)
{
	//1s
	clearTextBuf();
	StrCpyFromROM(gTextBuf, rManCheck, 13);
	lcd.setCursor(0, 0);
	lcd.print(gTextBuf);
	
	//Next row
	lcd.setCursor(0, 1);
	clearTextBuf();
	
	if( inNo == 0)
	{
	
		//2s
		StrCpyFromROM(gTextBuf, rCVEnterStart, 15);
		gTextBuf[0] = 0x3E;
		lcd.print(gTextBuf);
	
	}
	else if( inNo == 1)
	{
	
		//2s
		StrCpyFromROM(gTextBuf, rCVReading, 14);
		gTextBuf[0] = 0x3E;
		lcd.print(gTextBuf);
	
	}
	else if( inNo == 2)
	{
	
		//2s
		switch(Sequence.CVValue)
		{
		case 156:
			StrCpyFromROM(gTextBuf, rManId_1, sizeof(rManId_1));
			break;
		case 140:
			StrCpyFromROM(gTextBuf, rManId_2, sizeof(rManId_2));
			break;
		case 40:
			StrCpyFromROM(gTextBuf, rManId_3, sizeof(rManId_3));
			break;
		case 108:
			StrCpyFromROM(gTextBuf, rManId_4, sizeof(rManId_4));
			break;
		case 103:
			StrCpyFromROM(gTextBuf, rManId_5, sizeof(rManId_5));
			break;
		case 186:
			StrCpyFromROM(gTextBuf, rManId_6, sizeof(rManId_6));
			break;
		case 99:
			StrCpyFromROM(gTextBuf, rManId_7, sizeof(rManId_7));
			break;
		case 141:
			StrCpyFromROM(gTextBuf, rManId_8, sizeof(rManId_8));
			break;
		case 129:
			StrCpyFromROM(gTextBuf, rManId_9, sizeof(rManId_9));
			break;
		case 151:
			StrCpyFromROM(gTextBuf, rManId_10, sizeof(rManId_10));
			break;
		case 161:
			StrCpyFromROM(gTextBuf, rManId_11, sizeof(rManId_11));
			break;
		case 159:
			StrCpyFromROM(gTextBuf, rManId_12, sizeof(rManId_12));
			break;
		case 27:
			StrCpyFromROM(gTextBuf, rManId_13, sizeof(rManId_13));
			break;
		case 11:
			StrCpyFromROM(gTextBuf, rManId_14, sizeof(rManId_14));
			break;
		case 143:
			StrCpyFromROM(gTextBuf, rManId_15, sizeof(rManId_15));
			break;
		case 145:
			StrCpyFromROM(gTextBuf, rManId_16, sizeof(rManId_16));
			break;
		case 113:
			StrCpyFromROM(gTextBuf, rManId_17, sizeof(rManId_17));
			break;
		case 153:
			StrCpyFromROM(gTextBuf, rManId_18, sizeof(rManId_18));
			break;
		case 101:
			StrCpyFromROM(gTextBuf, rManId_19, sizeof(rManId_19));
			break;
		case 127:
			StrCpyFromROM(gTextBuf, rManId_20, sizeof(rManId_20));
			break;
		case 131:
			StrCpyFromROM(gTextBuf, rManId_21, sizeof(rManId_21));
			break;
		case 85:
			StrCpyFromROM(gTextBuf, rManId_22, sizeof(rManId_22));
			break;
		case 48:
			StrCpyFromROM(gTextBuf, rManId_23, sizeof(rManId_23));
			break;
		case 109:
			StrCpyFromROM(gTextBuf, rManId_24, sizeof(rManId_24));
			break;
		case 157:
			StrCpyFromROM(gTextBuf, rManId_25, sizeof(rManId_25));
			break;
		case 123:
			StrCpyFromROM(gTextBuf, rManId_26, sizeof(rManId_26));
			break;
		break;
		default:
			StrCpyFromROM(gTextBuf, rManId_0, 8);
		break;
		}
		
		for( int i = 1; i < 16; i++)
		{
			gTextBuf[(15 - i + 1)] = gTextBuf[15 - i];
		}
		
		gTextBuf[0] = 0x3E;
		
		
		lcd.print(gTextBuf);
	
	}
	else if (inNo == 3)
	{
		
		StrCpyFromROM(gTextBuf, rCVReadErr, 12);
		gTextBuf[0] = 0x3E;
		lcd.print(gTextBuf);
		
	}

	
}
//

void ExecuteCommand(unsigned char *outpReturn, unsigned char inDrawFlag)
{
	unsigned char aCVValue;
	
	PackedDScmd aDScmd = Sequence.GetCommad();
	
	if( inDrawFlag == 1)
	{
		Draw();
	}
	
	switch(aDScmd.mCmd)
	{
	case DSCMD_CVREAD:
		
		aCVValue = 0;
		
		if( DSCore.ReadConfig(ADDR_DCC, aDScmd.mParam1, &aCVValue, Sequence.gCfgReadSpeed, Sequence.gCfgReadPower))
		{
			//Success
			Sequence.CVValue = aCVValue;
			*outpReturn = 2;
			
		}
		else
		{
			//Error
			Sequence.CVValue = 0;
			*outpReturn = 3;
			
		}
		break;
		
	case DSCMD_CVWRITE:
		if( DSCore.WriteConfig_Dir( aDScmd.mParam1, aDScmd.mParam2))
		{
			*outpReturn = 3;
		}
		else
		{
			//Error
			*outpReturn = 4;
		}
			
		break;
	/*
	case DSCMD_SPEED:
	case DSCMD_DIR:
		break;
	case DSCMD_NONE:
		break;
	*/
	}
	
}


void changedTargetSpeed(const word inAddr, const word inSpeed)
{
	byte aProtocol = 2;
	
	DSCore.SetLocoSpeedEx(ADDR_DCC + inAddr, inSpeed * 100, aProtocol);
}

void changedTargetFunc(const word inAddr, const byte inFuncNo, const byte inFuncPower)
{
	DSCore.SetLocoFunction(ADDR_DCC + inAddr, inFuncNo, inFuncPower);
}

void changedTargetDir(const word inAddr, const byte inDir)
{
	DSCore.SetLocoDirection(ADDR_DCC + inAddr, inDir);
}

void changedTurnout(const word inAddr, const byte inDir)
{
	DSCore.SetTurnout(ADDR_ACC_DCC + inAddr - 1, inDir);
	
}

void changedPowerStatus(const byte inPower)
{
	DSCore.SetPower(inPower);
}

void StrCpyFromROM(char *outpText, const char *inpROMText, byte inLen)
{
	byte i;
	
	for( i = 0; i < inLen; i++)
	{
		outpText[i] = (char)pgm_read_byte(inpROMText + i);
	}
	
	for( i = inLen; i < 16; i++)
	{
		outpText[i] = 0x00;
	}
}